#define _GNU_SOURCE
#include <sched.h>
#include <linux/sched.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <linux/videodev2.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <semaphore.h>
#include <string.h>
#include <pthread.h>
#include <errno.h>
#include <sys/auxv.h>

#define MAP_SIZE 0x1c2000

void printBytes(unsigned char *address, unsigned size) {
    char buf[0x1000] = {0};
    unsigned count;
    for (count = 0; count < size; count++){
        if(!(count & 0b111111)) //every 64 bytes
            puts("");
        printf("%.2x ", address[count]);
    }
    puts("");
}
void printFirstBytes(unsigned char * address,unsigned size)
{
    for(unsigned i = 0; i < size; i += 0x1000)
    {
        printf("First bytes of page %u",i >> 12);
        printBytes(address,64);
    }
}
// This is unnecessary to functionize but it just feels nicer to me.
static inline char fault_page(void *addr, unsigned page)
{
    return ((char *)addr)[page * 0x1000];
}

int main()
{
    // Prep the videobuffer driver
    int vdfd = open("/dev/video0", O_RDWR);
    if (vdfd == -1)
    {
        printf("open: %m\n");
        return 1;
    }
    struct v4l2_requestbuffers rb;
    rb.count = 1;
    rb.memory = V4L2_MEMORY_USERPTR;
    rb.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    int ret = ioctl(vdfd, VIDIOC_REQBUFS, &rb);
    if (ret == -1)
    {
        printf("ioctl: %m\n");
        return 1;
    }
    int usbfd = open("/dev/bus/usb/001/001", O_RDONLY);
    if (usbfd == -1)
    {
        printf("open usb dev: %m\n");
        exit(1);
    }
    void *usbmap = mmap(NULL, MAP_SIZE, PROT_READ, MAP_PRIVATE, usbfd, 0);
    if (usbmap == (void *)-1)
    {
        printf("usb mmap: %m\n");
        exit(1);
    }
    printf("mapped usb memory at %p\n", usbmap);
    printFirstBytes(usbmap,MAP_SIZE >> 2);

    struct v4l2_buffer buf;
    memset(&buf, 0, sizeof(buf));
    buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    buf.length = MAP_SIZE;
    buf.bytesused = MAP_SIZE;
    buf.memory = V4L2_MEMORY_USERPTR;
    buf.m.userptr = (unsigned long)usbmap;
    buf.index = 0;
    ret = ioctl(vdfd, VIDIOC_QBUF, &buf);
    if (ret == -1)
    {
        printf("ioctl qbuf: %m\n");
        exit(1);
    }
    puts("streamon");
    fflush(stdout);
    ret = ioctl(vdfd,VIDIOC_STREAMON,&buf.type);
    if(ret == -1) {
        printf("ioctl streamon: %m\n");
        exit(1);
    }
    sleep(1);

    close(vdfd);

    printFirstBytes(usbmap,MAP_SIZE >> 2);
    exit(0);
}